/*******************************************************************************
 *
 * File Name: neuralNetworkUpdateAlgorithm.cpp
 * Author: Richard Baxter
 * Project: Advanced Neural Network (ANN) - Fractal Based Neural Network
 * Project Version: 1b13d 29-Sept-07
 * Comments:
 *
 *******************************************************************************/

 //IMPORTANT CODING NOTE - on 10-dec-06/1b6b I have started to remove the neuronReference class as circular referencing does not appear to be allowed in C++
//NB when create NeuronList class change all referecnes to "...List->frontNeuronReferences.." to "...List->neuronReferences" as Neurons when declared for the purpose of creating neuron lists only have their frontNeuronReferences used


#include "neuralNetworkUpdateAlgorithm.h"
#include "neuronClass.h"

#include <ctime>
#include <cstdlib>	//for randon number generation
#include <cmath>

#include <string>
#include <iostream>
#include <fstream>
using namespace std;

#define	MU 0.1F


//this is a test .c file in creation
// relies on forward and backward relationships between neurons. each neuron has a number of 'front' and 'back' adjacent neuro

//static functions

static void forwardPassStep(Neuron * neuronBeingAccessed);
	static void adjustOutputValueOfANeuronBasedOnBackNeurons(Neuron * neuronBeingAccessed);

static void backwardPassStep(Neuron * neuronBeingAccessed, int isOutputLayer);
	static void calculateOutputErrorOfOutputNeuron(Neuron * neuronBeingAccessed);
	static void calculateOutputErrorOfNonoutputNeuron(Neuron * neuronBeingAccessed);
	static void calculateNewBackWeightOfNeuron(Neuron * neuronBeingAccessed);
	static void calculateNewBiasOfNeuron(Neuron * neuronBeingAccessed);



/************************************************************ Main Neural Network Learning Routines with a class Target set ******************************************************/
//preconditions; neural network is already filled with all necessary weights/biases/inputneuron values






//NB this generalised C++ algorithm is not the same as the original assignment algorithm, the difference being there are no distinct layers in the network, and so multiple forwards propogation occurs in a single forwards pass (for every neuron and its back and front relations).
//The alternative generalised C++ method (in order to conform to assignment) involves (based upon a 1D list list of input neurons) generating a list of neurons that represents the next layer and so forth.







double neuralNetworkBackPropogationPass(Neuron * firstInputNeuronInNetwork, Neuron * firstOutputNeuronInNetwork)
{
	forwardPassStep(firstInputNeuronInNetwork);

	backwardPassStep(firstOutputNeuronInNetwork, true);


	//basic forwards pass step algorithm (from assignment) to mirror in C++
	/*
	//forward pass

	//calaculates outputs1 (hidden Neuron outputs);
	for(i = 0; i < numberOfHiddenNeurons; i++)
	{
		float netI = 0.0F;
		for(j=0;j<numberOfInputNeurons;j++)
		{
			#ifdef DEBUG
			cout << "outputs0[" << j << "] = " << outputs0[j] << "\n" << endl;
			cout << "weights1[" << j*numberOfHiddenNeurons+i << "] = " << weights1[j*numberOfHiddenNeurons+i] << "\n" << endl;
			#endif

			netI = netI + (outputs0[j]*weights1[j*numberOfHiddenNeurons+i]);

		}

		netI = netI + biases1[i];
		outputs1[i] = calculateOValue(netI);	//set hidden neurons output

		#ifdef DEBUG
		cout << "netI " << i << " = " << netI << "\n" << endl;
		cout << "biases1[" << i << "] = " << biases1[i] << "\n" << endl;
		cout << "outputs1[" << i << "] = " << outputs1[i] << "\n" << endl;
		//error checking exit
		//exit(0);
		#endif

	}
	*/






	/*

	Error for signle training example over all output neurons:

	E = 1/2 SUM(0->i) e(i)^2 = 1/2 SUM (d(i) - o(i))^2

	therefore...

	*/

	double total2xError = 0.0F;

	Neuron * currentNeuronReference = firstOutputNeuronInNetwork;
	while(currentNeuronReference -> nextNeuron != NULL)
	{

		total2xError = total2xError + pow((currentNeuronReference->classTarget - currentNeuronReference->output), 2.0);

		/*total2xError = total2xError + pow((classTarget[i] - outputs2[i]),2);*/

		#ifdef DEBUG
		cout << "\nNeural Network Total Error Calculations" << endl;
		cout << "Neuron ID = " << currentNeuronReference->id << endl;
		cout << "Neuron classTarget = " << currentNeuronReference->classTarget << endl;
		cout << "Neuron Output = " << currentNeuronReference->output << endl;
		cout << "total2xError = " << total2xError << "\n" << endl;
		#endif

		currentNeuronReference = currentNeuronReference->nextNeuron;
	}

	float totalError = 0.5F * total2xError;

	#ifdef DEBUG
	cout << "total error = " << totalError << "\n" << endl;
	//error checking exit
	//exit(0);
	#endif


	//basic total error calculation (from assignment) to mirror in C++
	/*
	float total2xError = 0;
	for(i = 0; i < numberOfOutputNeurons; i++)
	{

		total2xError = total2xError + pow((classTarget[i] - outputs2[i]),2);
		#ifdef DEBUG
		cout << "classTarget[" << i << "] = " << classTarget[i] << "\n" << endl;
		cout << "outputs2[" << i << "] = " << outputs2[i] << "\n" << endl;
		cout << "total2xError = " << total2xError << "\n" << endl;
		#endif
	}

	float totalError = 0.5F * total2xError;

	#ifdef DEBUG
	cout << "total error = " << totalError << "\n" << endl;
	//error checking exit
	//exit(0);
	#endif
	*/


	return totalError;
}
/************************************************************ End Main Neural Network Learning Routines with a class Target set ***************************************************/





#ifdef ANN

void copyNeuronReferenceArray(Neuron * neuronReferenceArrayToUpdate[], Neuron * neuronReferenceArrayToCopy[], long numberOfNeuronsInReferenceList)
{
	Neuron * currentNeuronReference;
	long q;
	for(q = 0; q < numberOfNeuronsInReferenceList; q++)
	{
		copyNeuronKeyProperties(neuronReferenceArrayToUpdate[q], neuronReferenceArrayToCopy[q]);
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}
}

void copyNeuronReferenceArrayIntoNeuronList(Neuron * firstNeuronInListToUpdate, Neuron * neuronReferenceArrayToCopy[])
{
	Neuron * currentNeuronReference = firstNeuronInListToUpdate;
	long q;
	for(q = 0; q < currentNeuronReference->numNeuronsInBackLayerOfSubnet; q++)
	{
		copyNeuronKeyProperties(currentNeuronReference, neuronReferenceArrayToCopy[q]);
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}
}

void copyNeuronListIntoNeuronReferenceArray(Neuron * neuronReferenceArrayToUpdate[], Neuron * firstNeuronInListToCopy)
{
	Neuron * currentNeuronReference = firstNeuronInListToCopy;
	long q;
	for(q = 0; q < currentNeuronReference->numNeuronsInBackLayerOfSubnet; q++)
	{
		copyNeuronKeyProperties(neuronReferenceArrayToUpdate[q], currentNeuronReference);
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}
}

#endif


/************************************************************ Forward Pass Step Routines ******************************************************/

static void forwardPassStep(Neuron * neuronBeingAccessed)
{
	Neuron * currentNeuronReference = neuronBeingAccessed->firstNeuronInFrontLayer;

	while(currentNeuronReference -> nextNeuron != NULL)
	{
		#ifdef DEBUG
		cout << "\nA Hidden level Input Neuron has been selected for Learning" << endl;
		cout << "Neuron ID = " << currentNeuronReference->number << endl;
		cout << "Neuron Output = " << currentNeuronReference->output << endl;
		#endif

		adjustOutputValueOfANeuronBasedOnBackNeurons(currentNeuronReference);

		currentNeuronReference = currentNeuronReference->nextNeuron;
	}

	//ANN recursion
	#ifdef ANN
	Neuron * currentNeuronReference = neuronBeingAccessed;
	while(currentNeuronReference -> nextNeuron != NULL)
	{
		if(currentNeuronReference->isSubnet)
		{
				copyNeuronReferenceArrayIntoNeuronList(currentNeuronReference->firstNeuronInBackLayerOfSubnet, currentNeuronReference->backNeuronReferences);
				copyNeuronReferenceArrayIntoNeuronList(currentNeuronReference->firstNeuronInFrontLayerOfSubnet, currentNeuronReference->frontNeuronReferences);

			forwardPassStep(currentNeuronReference->firstNeuronInBackLayerOfSubnet->firstNeuronInFrontLayer);

				copyNeuronListIntoNeuronReferenceArray(currentNeuronReference->backNeuronReferences, currentNeuronReference->firstNeuronInBackLayerOfSubnet);
				copyNeuronListIntoNeuronReferenceArray(currentNeuronReference->frontNeuronReferences, currentNeuronReference->firstNeuronInFrontLayerOfSubnet);

		}
		currentNeuronReference = currentNeuronReference->nextNeuron;

	}
	#endif


	//recursion
	if(currentNeuronReference->hasFrontLayer)
	{
		forwardPassStep(neuronBeingAccessed->firstNeuronInFrontLayer);
	}


}


//Only [top level]/[foward level]/[final] Output Neurons have "Class Target" values
static void adjustOutputValueOfANeuronBasedOnBackNeurons(Neuron * neuronBeingAccessed)
{

	//for every Neuron being accessed

	float netI = 0.0F;
	for(long i = 0; i < neuronBeingAccessed->numBackNeuronConnections; i++)
	{
		netI = netI + (neuronBeingAccessed->backNeuronReferences[i]->output * neuronBeingAccessed->backNeuronRefConnections[i]->weight);
		/*netI = netI + (outputs0[j]*weights1[j*numberOfHiddenNeurons+i]);*/
	}

	netI = netI + neuronBeingAccessed->bias;		//this may not be necessary or meaningful
	/*netI = netI + biases1[i];*/

		//neuronBeingAccessed->netI = netI; << old
	neuronBeingAccessed->output = calculateOValue(netI);	//set ("hidden") Neuron output value

	#ifdef DEBUG
	cout << "\nNeuron Output Value Adjusted" << endl;
	cout << "Neuron ID = " << neuronBeingAccessed->id << endl;
	cout << "Neuron netI = " << netI << endl;		//this may not be necessary or meaningful
	cout << "Neuron Bias = " << neuronBeingAccessed->bias << endl;
	cout << "Neuron Output = " << neuronBeingAccessed->output << endl;
	#endif
}

/************************************************************ End Forward Pass Step Routines **************************************************/
















/************************************************************ Backward Pass Step Routines ******************************************************/


//preconditions - assumes neuronBeingAccessed has a back neuron references. (ie neuronBeingAccessed->firstBackNeuronReference != NULL)
static void backwardPassStep(Neuron * neuronBeingAccessed, int isOutputLayer)
{

	if(isOutputLayer)
	{	//calculate errors of output neurons - errors2 (output errors)

		//NB calculate error only for the neuron being Accessed [and not for an array of neurons at the same level)]
		Neuron * currentNeuronReference = neuronBeingAccessed;
		while(currentNeuronReference -> nextNeuron != NULL)
		{
			calculateOutputErrorOfOutputNeuron(neuronBeingAccessed);
			currentNeuronReference = currentNeuronReference->nextNeuron;
		}
	}
	else
	{//calculates the errors of the hidden neurons

		Neuron * currentNeuronReference = neuronBeingAccessed;
		while(currentNeuronReference -> nextNeuron != NULL)
		{
			calculateOutputErrorOfNonoutputNeuron(neuronBeingAccessed);
			currentNeuronReference = currentNeuronReference->nextNeuron;
		}
	}

	//NB calculate back Weight only for the neuron being Accessed [and not for an array of neurons at the same level)]
	Neuron * currentNeuronReference = neuronBeingAccessed;
	while(currentNeuronReference -> nextNeuron != NULL)
	{
		calculateNewBackWeightOfNeuron(neuronBeingAccessed);
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}


	//NB calculate bias only for the neuron being Accessed [and not for an array of neurons at the same level)]
	currentNeuronReference = neuronBeingAccessed;
	while(currentNeuronReference -> nextNeuron != NULL)
	{
		calculateNewBiasOfNeuron(neuronBeingAccessed);
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}

	//ANN recursion
#ifdef ANN
	Neuron * currentNeuronReference = neuronBeingAccessed;
	while(currentNeuronReference -> nextNeuron != NULL)
	{
		if(currentNeuronReference->isSubnet)
		{
				copyNeuronReferenceArrayIntoNeuronList(currentNeuronReference->firstNeuronInBackLayerOfSubnet, currentNeuronReference->backNeuronReferences);
				copyNeuronReferenceArrayIntoNeuronList(currentNeuronReference->firstNeuronInFrontLayerOfSubnet, currentNeuronReference->frontNeuronReferences);

			backwardPassStep(currentNeuronReference->firstNeuronInBackLayerOfSubnet, false);

				copyNeuronListIntoNeuronReferenceArray(currentNeuronReference->backNeuronReferences, currentNeuronReference->firstNeuronInBackLayerOfSubnet);
				copyNeuronListIntoNeuronReferenceArray(currentNeuronReference->frontNeuronReferences, currentNeuronReference->firstNeuronInFrontLayerOfSubnet);
		}
		currentNeuronReference = currentNeuronReference->nextNeuron;
	}
#endif

	//recursion
	if(currentNeuronReference->hasBackLayer)
	{
		backwardPassStep(neuronBeingAccessed->firstNeuronInBackLayer, false);
	}





	//basic backward pass step algorithm (from assignment) to mirror in C++
	/*
	if(isOutputLayer)
	{
		//calculate errors of output neurons - errors2 (output errors)

		for(i = 0; i < numberOfOutputNeurons; i++)
		{

			float outputError = (classTarget[i] - outputs2[i]) * outputs2[i] * (1 - outputs2[i]);
			errors2[i] = outputError;

			#ifdef DEBUG
			cout << "classTarget[" << i << "] = " << classTarget[i] << "\n" << endl;
			cout << "outputs2[" << i << "] = " << outputs2[i] << "\n" << endl;
			cout << "errors2[" << i << "] = " << errors2[i] << "\n" << endl;
			#endif
		}
	}
	else
	{
		for(i = 0; i < numberOfHiddenNeurons; i++)
		{
			float hiddenError;

			float partOfHiddenError = 0;	//eg error3 = o3*(1-03)*partOfHiddenError where partOfHiddenError = (w36*o6 +w37*o7)
			for(j = 0; j < numberOfOutputNeurons; j++)
			{
				#ifdef DEBUG
				cout << "weights2[" << i*numberOfOutputNeurons+j << "] = " << weights2[i*numberOfOutputNeurons+j] << "\n" << endl;
				cout << "errors2[" << j << "] = " << errors2[j] << "\n" << endl;
				cout << "partOfHiddenError = " << partOfHiddenError << "\n" << endl;
				#endif
				partOfHiddenError = partOfHiddenError + (weights2[i*numberOfOutputNeurons+j]*errors2[j]);
			}

			errors1[i] = outputs1[i]*(1-outputs1[i])*partOfHiddenError;

			#ifdef DEBUG
			cout << "partOfHiddenError = " << partOfHiddenError << "\n" << endl;
			cout << "outputs1[" << i << "] = " << outputs1[i] << "\n" << endl;
			cout << "errors1[" << i << "] = " << errors1[i] << "\n" << endl;
			#endif
		}
	}

	float mu = MU;

	//calculates new weights between the hidden and output neurons

	for(i = 0; i < numberOfHiddenNeurons; i++)
	{
		for(j = 0; j < numberOfOutputNeurons; j++)
		{
			float weightChange;
			weightChange = mu * errors2[j] * outputs1[i];

			weights2[i* numberOfOutputNeurons+j] = weights2[i* numberOfOutputNeurons+j] + weightChange;

			#ifdef DEBUG
			cout << "mu = " << mu << "\n" << endl;
			cout << "errors2[" << j << "] = " << errors2[j] << "\n" << endl;
			cout << "outputs1[" << i << "] = " << outputs1[i] << "\n" << endl;
			cout << "weights2[" << i*numberOfOutputNeurons+j << "] = " << weights2[i*numberOfOutputNeurons+j] << "\n" << endl;
			#endif

		}
	}

	//calculates new biases of the output neurons

	for(i = 0; i < numberOfOutputNeurons; i++)
	{
		float biasChange;
		biasChange = mu * errors2[i] * 1.0F;	//BIASES ARE WEIGHTED WITH INPUT 1
		biases2[i] = biases2[i] + biasChange;

		#ifdef DEBUG
		cout << "errors2[" << i << "] = " << errors2[i] << "\n" << endl;
		cout << "biases2[" << i << "] = " << biases2[i] << "\n" << endl;
		#endif
	}
	*/

}



static void calculateOutputErrorOfOutputNeuron(Neuron * neuronBeingAccessed)
{
	double outputError = (neuronBeingAccessed->classTarget - neuronBeingAccessed->output) * neuronBeingAccessed->output * (1.0F - neuronBeingAccessed->output);
	neuronBeingAccessed->error = outputError;

	#ifdef DEBUG
	cout << "\nNeuron Output error Adjusted" << endl;
	cout << "Neuron ID = " << neuronBeingAccessed->id << endl;
	cout << "Neuron classTarget = " << neuronBeingAccessed->classTarget << endl;
	cout << "Neuron Error = " << neuronBeingAccessed->error << endl;
	cout << "Neuron Output = " << neuronBeingAccessed->output << endl;
	#endif

	/*
	float outputError = (classTarget[i] - outputs2[i]) * outputs2[i] * (1 - outputs2[i]);
	errors2[i] = outputError;
	*/

}

static void calculateOutputErrorOfNonoutputNeuron(Neuron * neuronBeingAccessed)
{
	float partOfHiddenError = 0;
	for(long i = 0; i < neuronBeingAccessed->numFrontNeuronConnections; i++)
	{
		#ifdef DEBUG
		cout << "\nNeuron Output error Adjusted" << endl;
		cout << "Neuron ID = " << neuronBeingAccessed->id << endl;
		cout << "Neuron Connection Weight = " << neuronBeingAccessed->frontNeuronRefConnections[i]->weight << endl;
		cout << "Neuron Error = " << neuronBeingAccessed->error << endl;
		cout << "partOfHiddenError = " << partOfHiddenError << "\n" << endl;
		#endif

		partOfHiddenError = partOfHiddenError + ((neuronBeingAccessed->frontNeuronRefConnections[i]->weight) * (neuronBeingAccessed->frontNeuronReferences[i]->error));
		/*partOfHiddenError = partOfHiddenError + (weights2[i*numberOfOutputNeurons+j]*errors2[j]);*/
	}

	neuronBeingAccessed->error = (neuronBeingAccessed->output) * (1.0 - (neuronBeingAccessed->output)) * partOfHiddenError;
	/*errors1[i] = outputs1[i]*(1-outputs1[i])*partOfHiddenError;*/

	#ifdef DEBUG
	cout << "Neuron ID = " << neuronBeingAccessed->id << endl;
	cout << "partOfHiddenError = " << partOfHiddenError << "\n" << endl;
	cout << "Neuron output  = " << neuronBeingAccessed->output << endl;
	cout << "Neuron Error = " << neuronBeingAccessed->error << endl;
	#endif
}


static void calculateNewBackWeightOfNeuron(Neuron * neuronBeingAccessed)
{
	for(long i = 0; i < neuronBeingAccessed->numBackNeuronConnections; i++)
	{
		float weightChange;
		weightChange = MU * neuronBeingAccessed->error * neuronBeingAccessed->backNeuronReferences[i]->output;
		neuronBeingAccessed->backNeuronRefConnections[i]->weight = neuronBeingAccessed->backNeuronRefConnections[i]->weight + weightChange;
	}

	#ifdef DEBUG
	cout << "\nNeuron Back Weight Value Adjusted" << endl;
	cout << "Neuron ID = " << neuronBeingAccessed->id << endl;
	cout << "mu = " << mu << endl;
	cout << "Neuron Error = " << neuronBeingAccessed->error << endl;
	cout << "Neuron Weight = " << neuronBeingAccessed->weight << endl;
	cout << "nNeuron Output = " << neuronBeingAccessed->output << endl;
	#endif

	/*
	float weightChange;
	weightChange = mu * errors2[j] * outputs1[i];

	weights2[i* numberOfOutputNeurons+j] = weights2[i* numberOfOutputNeurons+j] + weightChange;
	*/

}

static void calculateNewBiasOfNeuron(Neuron * neuronBeingAccessed)
{
	double biasChange;
	biasChange = MU * neuronBeingAccessed->error * 1.0F;	//BIASES ARE WEIGHTED WITH INPUT 1
	neuronBeingAccessed->bias = neuronBeingAccessed->bias + biasChange;

	#ifdef DEBUG
	cout << "\nNeuron Bias Value Adjusted" << endl;
	cout << "neuron ID = " << neuronBeingAccessed->id << endl;
	cout << "neuron Error = " << neuronBeingAccessed->error << endl;
	cout << "neuron Bias = " << neuronBeingAccessed->bias << endl;
	#endif

	/*
	float biasChange;
	biasChange = mu * errors2[i] * 1.0F;	//BIASES ARE WEIGHTED WITH INPUT 1
	biases2[i] = biases2[i] + biasChange;
	*/
}



/************************************************************ End Backward Pass Step Routines **************************************************/







float calculateOValue(float netValue)
{
	float f = (1.0/(1.0 + float(exp(-netValue))));	//modified 1-6-04
	return f;
}
















